/* ***************************************************************************
 *
 *                              KisSplice
 *      de-novo calling alternative splicing events from RNA-seq data.
 *
 * ***************************************************************************
 *
 * Copyright INRIA
 *  contributors :  Vincent Lacroix
 *                  Pierre Peterlongo
 *                  Gustavo Sacomoto
 *                  Vincent Miele
 *                  Alice Julien-Laferriere
 *                  David Parsons
 *
 * pierre.peterlongo@inria.fr
 * vincent.lacroix@univ-lyon1.fr
 *
 * This software is a computer program whose purpose is to detect alternative
 * splicing events from RNA-seq data.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".

 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.

 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <vector>
#include <stack>
#include <utility>
#include <algorithm>
#include "debug.h"
#include "NGraph.h"



using namespace std;

typedef vector< pair<int, char> > cycle_t;

struct arg_t
{
  NGraph* g;
  int num_bubbles;
  char* comment;
  FILE* seq_file;
  FILE* cycle_file;
  FILE* stream;
};

#define MAX_VAL 10000000 // For BellmanFord Algo, will be eq to infinite
//#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b)) 
#define MIN3(a,b,c) ((a) < (b) ? ((a) < (c) ? (a) : (c)) : ((b) < (c) ? (b) : (c)))
// ===========================================================================
//                             Global Variables
// ===========================================================================
//used for type of polymorphism (outputBubble)
char output_snps;
int UL_MAX;// Maximum upper and lower path bond
int LL_MAX;
int LL_MIN;
int MAX_BUBBLES;
int min_dist;
// Used and filled per bellmanFord function
int *dist;

///////////////////////////////// EDIT DISTANCE ////////////////////////////// 
/*!
 * \brief compute an edit distance between two vector
 * \param s1 first vector 1
 * \param l1 size of vector 1
 * \param s2 vector 2
 * \param l2 size of vector 2
 * \param size of the object constituing the vector
 * \param comparison function
 *
 * Computes Levenshtein distance between s1 and s2 .
 */
int edit_distance(const void *s1, size_t l1, const void *s2, size_t l2, size_t nmemb, int (*comp)(const void*, const void*)) 
{
  unsigned int i;
  unsigned int j;
  size_t len = (l1 + 1) * (l2 + 1);
  char *p1;
  char *p2;
  unsigned int d1;
  unsigned int d2;
  unsigned int d3;
  unsigned int *d;
  unsigned int *dp;
  unsigned int res;
 
  if (l1 == 0)  
    return l2;
  else if (l2 == 0) 
    return l1;
   
  d = (unsigned int*)malloc(len * sizeof(unsigned int));
 
  *d = 0;
  for(i = 1, dp = d + l2 + 1; i < l1 + 1; ++i, dp += l2 + 1)
  {
    *dp = (unsigned) i;
  }
  for(j = 1, dp = d + 1; j < l2 + 1; ++j, ++dp)
  {
    *dp = (unsigned) j;
  }

  for(i = 1, p1 = (char*) s1, dp = d + l2 + 2; i < l1 + 1; ++i, p1 += nmemb, ++dp)
  {
    for(j = 1, p2 = (char*) s2; j < l2 + 1; ++j, p2 += nmemb, ++dp) 
    {
      if(!comp(p1, p2))
      {
        *dp = *(dp - l2 - 2);// same base
      }
      else 
      {
        d1 = *(dp - 1) + 1;// insertion
        d2 = *(dp - l2 - 1) + 1;// deletion
        d3 = *(dp - l2 - 2) + 1;// substitution
        *dp = MIN3(d1, d2, d3);
      }
    }
  }
  res = *(dp - 2);
 
  dp = NULL;
  free(d);
  return res;
}

int hamming_distance(const void *s1, size_t l1, const void *s2, size_t l2, size_t nmemb, int (*comp)(const void*, const void*)) 
{
  char *p1, *p2;
 
  unsigned int i, res = 0;
  for (i = 0, p1 = (char *)s1, p2 = (char *)s2; i < l1; i++, p1 += nmemb, p2 += nmemb)
    if (comp(p1,p2))
      res++;

  return res;
}

int comp(const void *a, const void *b)
{
  return *(char *)a != *(char *)b;
}
/////////////////////////////// END EDIT DISTANCE ////////////////////////////

/*!
 * \brief Output the summary of the analysis.
 * Number of cycles and bubbles
 */
void printSummary( arg_t &arg )
{
  fprintf( arg.stream, "============================================================================\n" );
  fprintf( arg.stream, "Summary of results\n" );
  fprintf( arg.stream, "============================================================================\n" );
  
  fprintf( arg.stream, "No of bubbles: %d\n", arg.num_bubbles );

  fprintf( arg.stream, "============================================================================\n" );
}

/*!
\brief Used for outputting the bubble (in outputBubble)
returns the full sequence of the path from start to end in cycle
\param g the current graph
\param start
\param end
\param cycle
*/
string inducedPathInfo( const NGraph &g, const int start, const int end, const cycle_t &cycle )
{
  string res("");
  for (int i = (start + 1) % (int)cycle.size(); i != end; i = (i + 1) % (int)cycle.size() )
  {
    string curr = cycle[i].second == 'R' ? reverse_complement( g.getSequence(cycle[i].first) ) : g.getSequence( cycle[i].first );
    
    // It's not the first node, removes the 'k-1' prefix (already in the previous node)
    if ( (int)res.size() != 0 )
      curr.erase( 0, g.getKValue()-1 );
    
    res += curr;
  }
  
  return res;
}
/*
 * \brief Output the current bubble in the file opened in arg and classifies it
 * \param cycle: the cycle founded. used to get the paths sequences
 * \param sw_nodes: id  (int) of the switching nodes of the bubble (has a size 2)
 * \param arg: arg_t, the graph, file being written, cycle number...
 */
void outputBubble(const cycle_t &cycle, const vector<int> &sw_nodes, arg_t &arg )
{
  string lower_seq;
  string upper_seq;
  
  lower_seq = inducedPathInfo( *(arg.g), sw_nodes[0], sw_nodes[1], cycle );
  upper_seq = reverse_complement( inducedPathInfo( *(arg.g), sw_nodes[1], sw_nodes[0], cycle ) );

 
  if (upper_seq.size() < lower_seq.size())
    swap(upper_seq, lower_seq);

  if ((int)lower_seq.size() > LL_MAX || (int)lower_seq.size() < LL_MIN ||
      (int)upper_seq.size() > UL_MAX )
    return;
  
  ///------------ COMPUTE type (PIERRE, USING GUSTAVO's edit_distance function)
  int type = 0;
  int u_len = (int)upper_seq.size();
  int l_len = (int)lower_seq.size();
  
  if ( (u_len - l_len > 0) &&  (u_len - l_len) <= 2)   // (indel)
    type = 3;     
  else if ( l_len <= (2 * arg.g->getKValue() - 2) )           // (min path of at most 2k-2: splicing, indel, repeats)
  {

    int d1 = edit_distance(upper_seq.c_str(), l_len, lower_seq.c_str(), l_len, sizeof(char), comp);
    int d2 = edit_distance(upper_seq.c_str() + u_len - l_len, l_len, lower_seq.c_str(), l_len, sizeof(char), comp);

    if (d1 <= min_dist || d2 <= min_dist)   // (tandem repeat) 
      type = 2;
    else                                    // (splicing)
      type = 1; 
  }
  else if (u_len == l_len && // SNP, at most 10% differences. Arbitrary TODO: check this
	   ( hamming_distance(upper_seq.c_str(), l_len, lower_seq.c_str(), l_len, sizeof(char), comp) <= (0.1*l_len)) )
    type = 0;                       
  else // others with lower path of length bigger than 2k-2
    type = 4;
  
  /// END COMPUTE TYPE
//------ Outputting the bubble
  if(output_snps || type) // Dont output if type==0 and output_snps==0
  { 
    fprintf(arg.seq_file, ">%s|Cycle_%d|Type_%d|upper_path_Length_%d\n", arg.comment, arg.g->getNbOutput(), type, u_len);
    fprintf(arg.seq_file, "%s\n", upper_seq.c_str());
    
    fprintf(arg.seq_file, ">%s|Cycle_%d|Type_%d|lower_path_Length_%d\n", arg.comment, arg.g->getNbOutput(), type, l_len);
    fprintf(arg.seq_file, "%s\n", lower_seq.c_str());
  }

  arg.num_bubbles++ ;
  arg.g->incrementNbOutput() ;
  // Stop if the maximum number of bubbles was reached
  if (arg.num_bubbles >= MAX_BUBBLES)
  {
    // DON'T CHANGE THE RETURN CODE! It affects kissplice main script.
    exit(15);
  }  
}

/*\brief Update the directions of 'cycle', with regard to edge i -> j (label) and the switching nodes.
 * cycle only contains the number (id, an int) of the nodes to follow, its directions
 * are still initialized to "-". Directions will be changed for cycle[i] and cycle[j].
 * \param i: node i
 * \param j: node j
 * \param label: label of the edge i->j
 * \param cycle: cycle followed (a vector<pair<int, char>>)
 * \param sw_nodes: switching nodes for the cycle being treated
 */
void updateSw( const int i, const int j, const string label, cycle_t &cycle, vector<int> &sw_nodes)
{
  if (cycle[i].second != '-' && cycle[i].second != label[0])
  {// directions were already updated from an other edge: ie we reach a branching node, a switching node
    sw_nodes.push_back(i);
  }
  // The direction orientation of cycle[i] should _always_ be coherent
  // with the out-going arc i -> j
  cycle[i].second = label[0];
  
  if (cycle[j].second != '-' && cycle[j].second != label[1])
    sw_nodes.push_back(j);
  else
    cycle[j].second = label[1];
}


/*!
 * \brief Output all bubbles present in cycle
 * \param start start position in cycle, always 0
 * \param cycle cycle found, a  vector<pair<int, char>>, int being node id , char direction of the edge
 * \param sw_nodes
 * \param arg
 */
void find_all_induced_bubbles( const int start, cycle_t &cycle, vector<int> &sw_nodes, arg_t &arg )
{
  for ( int i = start ; i < (int)cycle.size() ; i++ )
  {
    int next = (i + 1) % (int)cycle.size();// 0 when cycle size reached, i+1 otherwise
    int u = cycle[i].first;
    int v = cycle[next].first;
    
    list<NEdge>::const_iterator it;
    // for: to get iterator on u->v edge
    for ( it = arg.g->getAdjList(u).begin() ; it != arg.g->getAdjList(u).end() ; it++ )
    {
      if ( it->get_node() == v )// it in the correct edge
      {
        break;
      }
    }
   
    // Just one possible edge, no recursive call
    updateSw( i, next, it->get_labels(), cycle, sw_nodes );

    if ( (int) sw_nodes.size() > 2 )//? more than two branching nodes discovered?
    {
      return;
    }
  }
  
  // If not exactly 2 switching nodes.
  if ( (int) sw_nodes.size() != 2 )
  {
    return;
  }
 // if two switching nodes, a bubble
  outputBubble( cycle, sw_nodes, arg );
}

/*!
 * \brief Handles the cycle existing in point, outputting bubbles if present.
 * \param point: the stack of the followed nodes which lead to a cycle
 * \param arg: arg_t structure, contains the graph
 */
void handleCycle( stack<int> point, arg_t & arg )
{
  vector<int> sw_nodes;
  cycle_t cycle;
  
  while (!point.empty())
  {
    cycle.push_back(make_pair(point.top(), '-'));
    point.pop();
  }  
  reverse(cycle.begin(), cycle.end());

  if ((int)cycle.size() > 2)
  { 
    find_all_induced_bubbles( 0, cycle, sw_nodes, arg );
  } 

  
}

/*!
 * \brief Compute Bellman Ford algorithm;
 * Find the shortest paths in g from source . !! Modifies the global variable dist
 * \param g  the graph
 * \param src number of the starting node
 */
void bellmanFord( const NGraph & g, const  int src )
{
  // Initialization
  for ( int i = 0 ; i < (int)g.getNbNodes() ; i++ )
  {
    dist[i] = MAX_VAL; 
  }
  dist[src] = 0;

  for ( int j = 0 ; j < (int)g.getNbNodes() ; j++ )// Repeat |V| -1 times
  {
    // Going through the edges
    for ( int u = 0 ; u < (int)g.getNbNodes() ; u++ )
    {

      list<NEdge>::const_iterator it;

      for ( it = g.getAdjList(u).begin() ; it != g.getAdjList(u).end() ; it++ )
      {
        if ( it->get_node() >= src )// if src reachable from u throught v
        {
          // weight of an edge (u,v): size sequence ( - overlap )
          int novo = dist[u] + ( g.getSequence(it->get_node()).size() - g.getKValue() + 1 );
          
          if ( novo < dist[it->get_node()] )
          {
            // new distance for v: min( dist(v) ,dist (u) + weight(u,v) )
            dist[it->get_node()] = novo;
          }
        }
      }
    }
  }

}
/*!
 *\brief Recursive algorithm. Backtracking all cycles from a source s of a graph (in arg)
 *\brief will lead to output all the bubbles
* \param s source node
* \param v  node
* \param point
* \param mark  vector of boolean, initialized to FALSE
* \param dir direction followed (edges)
* \param v_label label of the edge just followed or ""
* \param sw_count: switching nodes count (int)
* \param sw_seqs: switching nodes sequence size (int)
* \param size: path size (int)
* \param open_path
* \param arg
*/
void cyclesBacktrack(const int s, int v, stack<int> &point, bool *mark, stack<char> &dir,
		      const string v_label, int sw_count, int sw_seqs, int size, int open_path, arg_t &arg )
{
  bool removed = false;

  // Number of SN prunning
  int new_sw = 0;
  if (dir.size() > 1) //edge was followed
  {
    char top_label = dir.top();
    if (top_label != v_label[0])
    { 
      if (sw_count == 2)// we passed the bubble
  	return;
      new_sw = 1;
      sw_seqs += arg.g->getSequence(point.top()).size() - arg.g->getKValue() + 1;
    }
  }
    
  //--------- Update paths size
  if (dir.size() != 0)// an edge was followed
  {
    size += arg.g->getSequence(v).size() - arg.g->getKValue() + 1;
  }

  if (new_sw && sw_count == 0)// a first switching node was find
  {
    open_path = arg.g->getSequence(v).size() + (arg.g->getSequence(point.top()).size() - arg.g->getKValue() + 1);
  }

  if (!new_sw && sw_count == 1)//path continue (no new switching node
  {
    open_path += arg.g->getSequence(v).size() - arg.g->getKValue() + 1;
  }
  //---------

  // We don't know if the last node is going to be switching or not.
  int last_node = (new_sw + sw_count) == 2 ? 0 : (arg.g->getSequence(v).size() - arg.g->getKValue() + 1);
  int lip_open = open_path - sw_seqs - last_node;
  int lip_rest = size - open_path + arg.g->getKValue() - 1 - last_node;

  //--------- Lower lip prunning
  // if shorter path does not respect length constraint return
  if (lip_open > LL_MAX && lip_rest > LL_MAX)
    return;
  if ((sw_count + new_sw) == 2 && lip_open < LL_MIN)
    return;

   //--------- Maximum size prunning
  // if longer path does not respect length constraint return
  if (lip_open > UL_MAX || lip_rest > UL_MAX || (size - sw_seqs - last_node) > (UL_MAX + LL_MAX))
    return;
  if ((sw_count + new_sw) == 2 && ((lip_rest - sw_seqs + dist[v]) > UL_MAX || (size - sw_seqs + dist[v]) > (UL_MAX + LL_MAX)))
    return;
   
  //--------- if not out of bound
  sw_count += new_sw;
  point.push(v);
  dir.push(v_label.size() == 2 ? v_label[1] : '-');
  mark[v] = true;
  
  
  // TODO : this code should be rehandled, I had to do dirty things to make it work with the new
  // class architecture

  //---------outgoing edges of v
  list<NEdge>::const_iterator it;
  for ( it = arg.g->getAdjList(v).begin() ; it != arg.g->getAdjList(v).end() ; removed ? it : it++ )
  {
    int w = it->get_node();
    removed = false;
      
    if ( w < s )
    {
      removed = true;
      it = arg.g->eraseAdjList( v, w );
    }
    
    else if ( w == s )// reaching the source-> cycle, handleCycle will find all induced bubble
    {
      handleCycle( point, arg );
    }
    else if ( not mark[w] )// not already seen node recursive backtrack
    {
      cyclesBacktrack( s, w, point, mark, dir, it->get_labels(), sw_count, sw_seqs, size, open_path, arg );
    }
  }
 
  mark[v] = false;
  point.pop();
  dir.pop();
}
/*!
 * \brief Find all cycles contained in the graph of arg
 * \param arg  arg_t  variable (graph, file opened, bubble counter...)
 */
void enumerateAllCycles( arg_t& arg )
{
  dist = new int[arg.g->getNbNodes()];
  
  bool *mark = new bool[arg.g->getNbNodes()];
  for ( int i = 0 ; i < (int)arg.g->getNbNodes() ; i++ )
  {
    mark[i] = false;
  }
  
  stack<int> point;
  stack<char> dir;

  for ( int s = 0 ; s < (int)arg.g->getNbNodes() ; s++ ) 
  {
    bellmanFord( *(arg.g), s );// compute  and fill dist for s
    cyclesBacktrack( s, s, point, mark, dir, "", 0, 0, arg.g->getKValue()-1, 0, arg );
  }
}


//
void openAndReadFile( char* filename, void (*read_file)(NGraph &, FILE *), NGraph& g )
{
  FILE *file = fopen(filename, "r");
  
  if (file != NULL)
  {
    read_file(g, file);
  }
  else
  {
    fprintf(stderr, "Problem opening %s!\n", filename);
    exit(0);
  }
  
  fclose(file);
}

/*!
 * \brief Exit function if wrong arguments. Print message
 */
void printUsageAndExit( char * name )
{
	fprintf( stderr, "Usage: %s edge_file node_file k_value seq_file edit_distance_threshold comment [-u UL_MAX] [-L LL_MAX] [-l LL_MIN] [-M MAX_BUBBLES] -s\n", name );
	fprintf( stderr, "\t [-u UL_MAX] Maximal length of the upper path of each bubble. Default: 500\n" );
	fprintf( stderr, "\t [-L LL_MAX] Maximal length of the lower path of each bubble. Default: 2k-2\n" );
	fprintf( stderr, "\t [-l LL_MIN] Minimal length of the lower path of each bubble. Default: 2k-6\n" );
	fprintf( stderr, "\t [-M MAX_BUBBLES] Stop the process after this number of cycles. Default: 100 Millions\n" );
	exit( EXIT_FAILURE );
}

int main( int argc, char ** argv )
{
  if ( argc < 7 )
  {  
    fprintf( stderr, "Wrong number of arguments!\n" );
    printUsageAndExit( argv[0] );
  }

  output_snps = 1; // by default, we output snps.


  NGraph g( atoi( argv[3] ), atoi( argv[7]) );// constructor: kvalue & nbSnpoutput in previous step
  arg_t arg;

  arg.stream = stdout;

  openAndReadFile( argv[1], &read_graph_edges, g );
  openAndReadFile( argv[2], &read_graph_nodes, g );

  arg.seq_file  = fopen( argv[4], "w" );
  min_dist      = atoi( argv[5] );
  arg.comment   = argv[6];
// ? why default?
  LL_MAX = 2 * g.getKValue() - 2;
  LL_MIN = 2 * g.getKValue() - 6;
  UL_MAX = 500;
  MAX_BUBBLES = 100000000;

  // TODO : should be while temoin = getopt(...)
  // voir aussi l'orga du code pr�traitement...
  while ( true ) 
  {
    int temoin = getopt ( argc-6, &argv[6], "u:L:l:M:s" );
    if ( temoin == -1 )
    {
      break;
    }
    
    switch ( temoin )
    {
      case 'u' :
      {
        UL_MAX = atoi( optarg );
        break;
      }
      case 'L' :
      {
        LL_MAX = atoi( optarg );
        break;
      }
      case 'l' :
      {
        LL_MIN = atoi( optarg );
        break;
      }     
      case 'M':
      {
        MAX_BUBBLES = atoi( optarg );
        break;
      }
      case 's':
      {
        output_snps = 0;
        printf( "Will not output SNPs and sequencing errors\n" );
        break;
      }
      default:
      {
        printUsageAndExit( argv[0] );
      }
    }
  }

  g.expand_parallel_edges();
  
  arg.g = &g;
  arg.num_bubbles = 0;
  

  enumerateAllCycles( arg );
  printSummary( arg );
  fclose( arg.seq_file );
  
  return EXIT_SUCCESS;
}
