/* ***************************************************************************
 *
 *                              KisSplice
 *      de-novo calling alternative splicing events from RNA-seq data.
 *
 * ***************************************************************************
 *
 * Copyright INRIA
 *  contributors :  Vincent Lacroix
 *                  Pierre Peterlongo
 *                  Gustavo Sacomoto
 *                  Vincent Miele
 *                  Alice Julien-Laferriere
 *                  David Parsons
 *
 * pierre.peterlongo@inria.fr
 * vincent.lacroix@univ-lyon1.fr
 *
 * This software is a computer program whose purpose is to detect alternative
 * splicing events from RNA-seq data.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".

 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.

 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */




#ifndef COMPRESSED_GRAPH_H
#define COMPRESSED_GRAPH_H

// ===========================================================================
//                               Include Libraries
// ===========================================================================
#include <stdio.h>
#include <string.h>
#include <vector>
#include <stack>
#include <algorithm>
#include <utility>




// ===========================================================================
//                             Include Project Files
// ===========================================================================
#include "CEdge.h"
#include "LabelledCEdge.h"




// ===========================================================================
//                              Class declarations
// ===========================================================================



// ===========================================================================
//                             Declare Used Namespaces
// ===========================================================================
using namespace std;





//! A class representing the compacted De-Bruijn graphs
/*!\
 * \brief Its light structure is used for the detection of bi-connected
 *  component in the graph.
 */

#define MAX 1024


class CGraph
{
  public:
    // =======================================================================
    //                               Constructors
    // =======================================================================

  /*!
   * \brief CGraph constructor.
   * Initializes the attributes according to the node numbers
   * Creates directly the CGraph object with its edges from allEdges
  */
    CGraph( int nbNodes, vector<LabelledCEdge> & allEdges, int kValue );

    // =======================================================================
    //                               Destructors
    // =======================================================================
    ~CGraph(void);


    // =======================================================================
    //                              Public Attributes
    // =======================================================================
    int k_value;

    // =======================================================================
    //                              Public Methods
    // =======================================================================
    inline int *get_adj_list(int node) const;
    inline int get_adj_list_sz(int node) const;
    inline int get_n_nodes(void) const;

    void insert_unique_edge_dir(int u, int v);
    void insert_unique_edge(int u, int v);
    void destroy_adj_list(void);
    void destroy(void);

  protected:
    // =======================================================================
    //                            Forbidden Constructors
    // =======================================================================

   CGraph(const CGraph &model)
    {
      printf("%s:%d: error: call to forbidden constructor of CGraph.\n",
          __FILE__, __LINE__);
      exit(EXIT_FAILURE);
    };

   CGraph( void )
   {
         printf("%s:%d: error: call to forbidden constructor of CGraph.\n",
             __FILE__, __LINE__);
         exit(EXIT_FAILURE);
       };

    // =======================================================================
    //                              Protected Methods
    // =======================================================================

    // =======================================================================
    //                             Protected Attributes
    // =======================================================================
    // NOTA: The neighborhood size is always <= 8

    //! Adjacency list
    /*!
     * adj_list is a vector of int tables containing the IDs of "outgoing neighbours"
     * i.e. neighbours v having a u -> v edge (u being the present node).
     * adj_list_sz contains the corresponding sizes
     */
    vector<int*>  adj_list;
    vector<int>   adj_list_sz;
};

// ===========================================================================
//                          Inline functions' definition
// ===========================================================================
int *CGraph::get_adj_list(int node) const
{
	return adj_list[node];
}

int CGraph::get_adj_list_sz(int node) const
{
	return adj_list_sz[node];
}

int CGraph::get_n_nodes(void) const
{
	return (int) adj_list.size();
}


// ===========================================================================
//                          Other functions
// ===========================================================================

void read_node_file(FILE *node_file, vector<char *>& seqs, int k_val);
void read_edge_file(FILE *edge_file, vector<LabelledCEdge>& edges);


#endif // COMPRESSED_GRAPH_H
