/* ***************************************************************************
 *
 *                              KisSplice
 *      de-novo calling alternative splicing events from RNA-seq data.
 *
 * ***************************************************************************
 *
 * Copyright INRIA
 *  contributors :  Vincent Lacroix
 *                  Pierre Peterlongo
 *                  Gustavo Sacomoto
 *                  Vincent Miele
 *                  Alice Julien-Laferriere
 *                  David Parsons
 *
 * pierre.peterlongo@inria.fr
 * vincent.lacroix@univ-lyon1.fr
 *
 * This software is a computer program whose purpose is to detect alternative
 * splicing events from RNA-seq data.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can  use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".

 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.

 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */



// ===========================================================================
//                               Include Libraries
// ===========================================================================
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <vector>


// ===========================================================================
//                             Include Project Files
// ===========================================================================
#include <debug.h>
#include <NGraph.h>
#include <CGraph.h>
#include <CycleCompression.h>
#include <SplitBcc.h>

#define MAX 1024


// ===========================================================================
//                         Define Miscellaneous Functions
// ===========================================================================

FILE* open_file( char* filename )
{
  FILE* file = fopen( filename, "r" );
  
  if ( file == NULL )
  {
    fprintf( stderr, "Problem opening %s!\n", filename );
    exit( EXIT_FAILURE );
  }
  
  return file;
}


void read_edges_and_nodes( char* edges_fname, char* nodes_fname, const int k,
        vector<LabelledCEdge>& allEdges, vector<char*>& seqs )
{
  FILE* edge_file = open_file(edges_fname);
  FILE* node_file = open_file(nodes_fname);
  
  read_edge_file( edge_file, allEdges );
  read_node_file( node_file, seqs, k );
  
  fclose( edge_file );
  fclose( node_file );
}

int main( int argc, char** argv )
{
  vector<char*> seqs;
  
  vector<LabelledCEdge> allEdges;

  string base_name = "./bcc/graph";

  if ( argc < 4 )
  { 
    fprintf( stderr, "Wrong number of arguments!\n" );
    fprintf( stderr, "Usage: ./run_modules edge_file node_file k_value path_to_output\n" );
    return 0;
  }
  
  if ( argc == 5 )
  {
    base_name = argv[4];
  }
  
 // CGraph graph;
  int k_value = atoi( argv[3] );

  // Read input files
  read_edges_and_nodes( argv[1], argv[2], k_value, allEdges, seqs );
  
 // Creating & Initializing the graph with the edges reads
  CGraph graph( (int)seqs.size(), allEdges, k_value );

  // Decompose graph into BCCs
  fprintf(stdout, "Searching biconnected components...\n");
  vector< vector<CEdge> > bcc = find_bcc(graph);
  fprintf( stdout, "Number of biconnected components found: %d\n\n", (int)bcc.size() );

  
  graph.destroy_adj_list();
  NGraph* component;
  // For each BCC, ...
  for ( int i = 0 ; i < (int)bcc.size() ; i++ )
  {
    fprintf(stdout, "Processing component %d...\n", i+1);
    
    
    // Build component's output file names
    char edge_fname[1024];
    char node_fname[1024];
    char cycle_fname[1024];
    char log_fname[1024];
    sprintf( edge_fname,  "%s_edges_bcc_%d",  base_name.c_str(), i+1 );
    sprintf( node_fname,  "%s_nodes_bcc_%d",  base_name.c_str(), i+1 );
    sprintf( cycle_fname, "%s_cycles_bcc_%d", base_name.c_str(), i+1 );
    sprintf( log_fname,   "%s_log_bcc_%d",    base_name.c_str(), i+1 );
    
    
    // Build uncompact graph corresponding to current BCC
    component = new NGraph( graph, seqs, allEdges, bcc[i] );

    fprintf( stdout, "Initial size: %d nodes.\n", (int)component->getNbNodes() );

    // Compress linear paths of size > 2
    fprintf(stdout, "Compressing linear paths...\n");
    int original_size = (int)component->getNbNodes();
    component->compress_all_paths();

    fprintf( stdout, "Number of compressed nodes: %d.\n", original_size - (int)component->getNbNodes() );
    
    
    // Compress bubbles
    fprintf( stdout, "Compressing simple bubbles...\n" );
    FILE *log_file = fopen(log_fname, "w");
    if ( log_file == NULL )
    {
      fprintf( stderr, "file %s not writable, exit\n", log_fname );
      exit( EXIT_FAILURE );
    }
 
    int n_compressed_bubbles;
    component->compress_all_bubbles( &n_compressed_bubbles, log_file, i+1 );
    
    fprintf( stdout, "Number of compressed bubbles: %d.\n", n_compressed_bubbles );
    fclose( log_file );
    
    
    // If ... (TODO) write component's edges and nodes into output files
    if ( (int)component->getNbNodes() >= 4 )
    { 
      FILE *edge_file = fopen( edge_fname, "w"), *node_file = fopen( node_fname, "w");
    
      if ( edge_file == NULL )
      {
        fprintf( stderr, "file %s not writable, exit\n", edge_fname );
        exit( EXIT_FAILURE );
      }
      if ( node_file == NULL )
      {
        fprintf( stderr, "file %s not writable, exit\n", node_fname );
        exit( EXIT_FAILURE );
      }
      
      component->print_graph_edges( edge_file, NULL, NULL, false );
      component->print_graph_nodes( node_file, NULL, NULL, false );
      
      fclose( edge_file );
      fclose( node_file );
    }

    fprintf( stdout, "Final size: %d nodes.\n", (int)component->getNbNodes() );
    fprintf( stdout, "Done!\n\n" );
    
    delete component;
  } 
  
  
  // STL does not delete the objects in the vector when we call clear,
  // we have to do it manually !!!!!
  for ( vector<char*>::iterator it = seqs.begin() ; it != seqs.end() ; it++ )
  {
    delete [] *it;
  } 
  seqs.clear();
  
  return EXIT_SUCCESS;
}
